@@ -9,6 +9,13 @@ import psycopg2
 from psycopg2 import sql
 #import MySQLdb
 import pickle
+import math
+
+from google.cloud import bigquery
+from google.oauth2 import service_account
+
+
+import abc
 
 def read_file(file):
     """Reads txt file -> list"""
@@ -141,55 +148,81 @@ class Migrator:
         return(result)
         
         
-    def migration_list_to_json(self):
+    def migration_list_to_json(self, filename = None):
         result=json.dumps(self.migration_list)
         
-        with open("migrations//migration-"+str(self.migration_number)+".json","w+") as f:
-            f.write(result)
+        if filename is None or filename == "" or filename.isspace():
+            with open("migrations/migration-"+str(self.migration_number)+".json","w+") as f:
+                f.write(result)
+        else:
+            with open(f"{filename}.json","w+") as f:
+                f.write(result)
+
     def create_migrations_from_df(self,name, dataframe):
-        columns = list(dataframe.columns)
+        
+        columns, return_types = self.extract_columns_and_types_from_df(dataframe)
 
-        return_types = []
-        for col in dataframe:
-            t = dataframe.loc[0, col]
-            try:
-                return_types.append(type(t.item()).__name__)
-            except:
-                length = 2**( int(dataframe[col].str.len().max()) - 1).bit_length()
-                return_types.append('nvarchar(' + str(length) + ')' if  type(t).__name__ == 'str' else type(t).__name__)
-
-        #return_types = ['nvarchar(255)' if type == 'str' else type for type in return_types]
-        print(return_types)
-        if (columns[0] != "id"):
-            columns.insert(0, "id")
-            return_types.insert(0, "int")
         migration_dict = {"create": {"table_name": name, "columns": columns, "types": return_types}}
         self.migration_list.append(migration_dict)
         self.migration_list_to_json()
+        #return columns, return_types
+
+    def extract_columns_and_types_from_df(self, dataframe):
+        columns = list(dataframe.columns)
+
+        return_types = []
+        for column in dataframe:
+            t = dataframe.loc[0, column]
+            try:
+                return_types.append(type(t.item()).__name__)
+            except:
+                #length = 2**( int(dataframe[col].str.len().max()) - 1).bit_length()
+                length = int(dataframe[column].str.len().max())
+                length += 0.1*length
+                length = int(math.ceil(length/10.0))*10
+                return_types.append(f'nvarchar({length})' if  type(t).__name__ == 'str' else type(t).__name__)
+
+        if (columns[0] != "id"):
+            columns.insert(0, "id")
+            return_types.insert(0, "int")
+
         return columns, return_types
 
 
-
-
-class AbstractDB:
+class AbstractDB(abc.ABC):
     def __init__(self,config_file="config.ini",db_details=None):
         if db_details is None:    
             db_details=read_connection_details(config_file)
-        locally=True
+        self.locally=True
         if db_details["LOCALLY"]=="False":
-            locally=False
+            self.locally=False
 
         self.DB_SERVER=db_details["DB_SERVER"]
         self.DB_DATABASE=db_details["DB_DATABASE"]
         self.DB_USERNAME = db_details["DB_USERNAME"]
         self.DB_PASSWORD = db_details["DB_PASSWORD"]
+        if "DB_PORT" in db_details.keys():
+            self.DB_PORT = int(db_details["DB_PORT"])
+        else:
+            self.DB_PORT = None
         
         if "DB_DRIVER" in db_details.keys():
             self.DB_DRIVER = db_details["DB_DRIVER"]
         else:
             self.DB_DRIVER="ODBC Driver 13 for SQL Server"
-        
-        if locally:
+            
+        self.connect()
+    
+    @abc.abstractmethod
+    def connect_locally(self):
+        pass
+
+    @abc.abstractmethod
+    def connect_remotely(self):
+        pass    
+
+    def connect(self):
+        if self.locally:
             self.connect_locally()
         else:
             self.connect_remotely()
@@ -315,7 +348,10 @@ class Mysqldb(AbstractDB):
         print("DB connection established")
         
     def connect_remotely(self):
-        self.connection = MySQLdb.connect(host=self.DB_SERVER,user=self.DB_USERNAME,password=self.DB_PASSWORD,database=self.DB_DATABASE)
+        if self.DB_PORT is not None:
+            self.connection = MySQLdb.connect(host=self.DB_SERVER,port=self.DB_PORT,user=self.DB_USERNAME,password=self.DB_PASSWORD,database=self.DB_DATABASE)
+        else:    
+            self.connection = MySQLdb.connect(host=self.DB_SERVER,user=self.DB_USERNAME,password=self.DB_PASSWORD,database=self.DB_DATABASE)
         self.cursor = self.connection.cursor()
         print("DB connection established")
         
@@ -361,6 +397,58 @@ class PostgresDb(AbstractDBPostgres):
             table_dict[table]=PostgresTable.init_all_columns(self, table)
 
         return(table_dict)
+    
+    
+
+    
+class BigQueryDb:
+    def __init__(self,credentials_path,project_id):
+        #super().__init__(None,None)
+        self.credentials_path=credentials_path
+        self.project_id=project_id
+        
+        self.credentials = service_account.Credentials.from_service_account_file(self.credentials_path, scopes=["https://www.googleapis.com/auth/cloud-platform"],)
+
+        self.client = bigquery.Client(credentials=self.credentials, project=self.credentials.project_id,)
+
+
+        
+    def connect_remotely(self):
+        print("Connect remotely")
+        
+    def connect_locally(self):
+        print("Connect locally")
+        
+        
+    def execute(self,query):
+        
+        
+        # query_job = self.client.query(
+        #     """
+        #     SELECT
+        #       CONCAT(
+        #         'https://stackoverflow.com/questions/',
+        #         CAST(id as STRING)) as url,
+        #       view_count
+        #     FROM `bigquery-public-data.stackoverflow.posts_questions`
+        #     WHERE tags like '%google-bigquery%'
+        #     ORDER BY view_count DESC
+        #     LIMIT 10"""
+        # )
+
+
+        
+        query_job = self.client.query(query)
+    
+        
+        results = query_job.result()  # Waits for job to complete.
+        
+        for row in results:
+            print(row)
+            
+            #print("{} : {} views".format(row.id, row.link,row.title))
+        
+        
 
 class MongoDb(AbstractDBMongo):
     def connect_remotely(self):
@@ -466,7 +554,8 @@ class AbstractTable(AbstractJoinable):
     def __init__(self,db1,name,columns=None,types=None):
         super().__init__(db1,name,columns)
         self.types=types
-    @save_migration
+    
+    #@save_migration
     def drop(self):
         query="DROP TABLE "+self.name
         print(query)
@@ -482,6 +571,16 @@ class AbstractTable(AbstractJoinable):
 
     def insert_from_df(self,df,batch=1,try_mode=False):
         assert len(df.columns)+1==len(self.columns) #+1 because of id column
+
+        pd_nullable_dtypes = {pd.Int8Dtype(), pd.Int16Dtype(), pd.Int32Dtype(), pd.Int64Dtype(),
+                              pd.UInt8Dtype(), pd.UInt16Dtype(), pd.UInt32Dtype(), pd.UInt64Dtype(),
+                              pd.Float32Dtype(), pd.Float64Dtype()}
+        pd_nullable_columns = df.dtypes.isin(pd_nullable_dtypes)
+
+        # cast pd nullable columns to float - changing to 'NULL' then proceeds without an error.
+        # in database floats are again casted to integers where required
+        df = df.copy()
+        df.loc[:, pd_nullable_columns] = df.loc[:, pd_nullable_columns].astype(float)
         
         #handling nan values -> change to NULL TODO
         for column in list(df.columns):
@@ -846,6 +945,12 @@ class MysqlTable(MysqlSelectable,AbstractTable):
         types=temporary_table.get_all_types()
         return(cls(db1,name,columns,types))
         
+    
+    def drop(self):
+        query="DROP TABLE "+self.name+";"
+        print(query)
+        self.db1.execute(query)
+    
     #@save_migration #TODO: Uncomment
     def create(self,foreign_keys=None):
         assert len(self.columns)==len(self.types)
@@ -885,7 +990,7 @@ class MysqlTable(MysqlSelectable,AbstractTable):
                     if "int" in self.types[j+1]:
                         if replace_apostrophes:
                             rows[k][j]=str(rows[k][j]).replace("'","")
-                        query+="NULL"
+                        query+="NULL,"
                     else:
                         query+="NULL,"
                 elif "nvarchar" in self.types[j+1]:
